@@ -26,7 +26,7 @@ extern Config config;
 extern Sdl sdl;
 extern SDL_Surface *offscreen;
 extern SDL_Surface *bkgnd;
-extern int keystate[SDLK_LAST];
+extern SDL_Surface *previewpieces;
 extern Bowl *bowls[BOWL_COUNT];
 
 enum {
@@ -120,66 +120,6 @@ void bowl_compute_help_pos( Bowl *bowl )
     bowl->help_sy = j * bowl->block_size + bowl->sy;
 }
 
-/*
-====================================================================
-Compute position of preview
-====================================================================
-*/
-void bowl_compute_preview_pos( Bowl *bowl )
-{
-    int i, j;
-    int x1, y1, x2, y2; /* corners of the tile */
-    int w, h;
-    Block_Mask *bm = &block_masks[bowl->next_block_id];
-    if ( bowl->preview_center_sx == -1 ) 
-        return;
-    /* first tile */
-    x1 = y1 = -1;
-    for ( j = 0; j < 4; j++ ) {
-        for ( i = 0; i < 4; i++ ) {
-            if ( bm->mask[bm->rstart][i][j] ) {
-                y1 = j;
-                break;
-            }
-        }
-        if ( y1 != -1 ) break;
-    }
-    for ( i = 0; i < 4; i++ ) {
-        for ( j = 0; j < 4; j++ ) {
-            if ( bm->mask[bm->rstart][i][j] ) {
-                x1 = i;
-                break;
-            }
-        }
-        if ( x1 != -1 ) break;
-    }
-    /* last tile */
-    x2 = y2 = -1;
-    for ( j = 3; j >= 0; j-- ) {
-        for ( i = 3; i >= 0; i-- ) {
-            if ( bm->mask[bm->rstart][i][j] ) {
-                y2 = j;
-                break;
-            }
-        }
-        if ( y2 != -1 ) break;
-    }
-    for ( i = 3; i >= 0; i-- ) {
-        for ( j = 3; j >= 0; j-- ) {
-            if ( bm->mask[bm->rstart][i][j] ) {
-                x2 = i;
-                break;
-            }
-        }
-        if ( x2 != -1 ) break;
-    }
-    /* preview position */
-    w = (x2 - x1 + 1) * bowl->block_size;
-    h = (y2 - y1 + 1) * bowl->block_size;
-    bowl->preview_sx = bowl->preview_center_sx - x1*bowl->block_size - w/2;
-    bowl->preview_sy = bowl->preview_center_sy - y1*bowl->block_size - h/2;
-}
-
 /*
 ====================================================================
 Compute computer target
@@ -208,6 +148,21 @@ void bowl_compute_cpu_dest( Bowl *bowl )
     bowl->cpu_dest_score = cpu_data.dest_score;
 }
 
+/** Initialize bowl's block structure to hold piece with id at top of bowl
+ * and also compute position of shadow piece. */
+void bowl_init_current_piece(Bowl *bowl, int id) {
+	bowl->block.id = id;
+	bowl->block.x = 5 - block_masks[bowl->block.id].rx;
+	bowl->block.y = 0 - block_masks[bowl->block.id].ry;
+	bowl->block.sx = bowl->sx + bowl->block_size * bowl->block.x;
+	bowl->block.sy = bowl->sy + bowl->block_size * bowl->block.y;
+	bowl->block.rot_id = block_masks[bowl->block.id].rstart;
+	bowl->block.sw = 4 * bowl->block_size;
+	bowl->block.sh = 4 * bowl->block_size;
+	bowl->block.cur_y = bowl->block.y * bowl->block_size;
+	bowl_compute_help_pos(bowl);
+}
+
 /*
 ====================================================================
 Initiate next block. Set bowl::block::id to preview and get id of next
@@ -227,7 +182,8 @@ void bowl_select_next_block( Bowl *bowl )
 	bowl->next_block_id = next_blocks[bowl->next_blocks_pos++];
 	DPRINTF("Bowl %p: Current = %d, Next = %d\n", bowl, bowl->block.id, 
 							bowl->next_block_id);
-	if ( bowl->next_blocks_pos == next_blocks_size ) {
+	/* get new blocks before they run out so three piece preview will not crash */
+	if ( bowl->next_blocks_pos >= next_blocks_size - 3 ) {
 		DPRINTF("Need to fill new tetrominoes bags\n");
 		/* we have to keep part of buffer which other bowls 
 		 * might still need to go through */
@@ -262,15 +218,7 @@ void bowl_select_next_block( Bowl *bowl )
 	}
     
 	/* init rest of block structure */
-	bowl->block.x = 5 - block_masks[bowl->block.id].rx;
-	bowl->block.y = 0 - block_masks[bowl->block.id].ry;
-	bowl->block.sx = bowl->sx + bowl->block_size * bowl->block.x;
-	bowl->block.sy = bowl->sy + bowl->block_size * bowl->block.y;
-	bowl->block.rot_id = block_masks[bowl->block.id].rstart;
-	bowl->block.sw = bowl->block.sh = 4 * bowl->block_size;
-	bowl->block.cur_y = bowl->block.y * bowl->block_size;
-	bowl_compute_help_pos( bowl );
-	bowl_compute_preview_pos( bowl );
+	bowl_init_current_piece(bowl, bowl->block.id);
 
 	/* count i pieces for drought and stats */
 	if (bowl->block.id == 6) {
@@ -287,7 +235,7 @@ void bowl_select_next_block( Bowl *bowl )
 		bowl->drought++;
 
 	/* if CPU is in control get destination row & other stuff */
-	if ( !bowl->controls ) {
+	if (bowl->cpu_player) {
 		/* destination */
 		bowl_compute_cpu_dest( bowl );
 		
@@ -544,7 +492,7 @@ void bowl_add_score(Bowl *bowl, int lc)
 	score = base[lc] * (bowl->level + 1);
 
 	/* add 20% if no preview */
-	if ( !config.preview || bowl->preview_center_sx == -1 )
+	if (!bowl->preview)
 		score += 20 * score / 100;
 
 	counter_add(&bowl->score, score);
@@ -764,12 +712,10 @@ Happens next in update().
 */
 void bowl_drop_block( Bowl *bowl )
 {
-    while (bowl_check_piece_position(bowl,
-		    bowl->block.x, bowl->block.y+1,
-		    bowl->block.rot_id) == POSVALID) {
-        bowl->block.cur_y += bowl->block_size;
-        bowl->block.y = (int)bowl->block.cur_y / bowl->block_size;
-    }
+	while (bowl_piece_can_drop(bowl)) {
+		bowl->block.y++;
+		bowl->block.cur_y = bowl->block.y * bowl->block_size;
+	}
 }
 
 /*
@@ -779,7 +725,9 @@ Return True if CPU may drop/move down.
 */
 int bowl_cpu_may_move_down( Bowl *bowl )
 {
-    if ( !bowl->controls && bowl->cpu_down && bowl->block.x == bowl->cpu_dest_x && bowl->block.rot_id == bowl->cpu_dest_rot )
+    if ( bowl->cpu_player && bowl->cpu_down &&
+		    bowl->block.x == bowl->cpu_dest_x &&
+		    bowl->block.rot_id == bowl->cpu_dest_rot )
         return 1;
     return 0;
 }
@@ -1016,7 +964,11 @@ Create a bowl at screen position x,y. Measurements are the same for
 all bowls. Controls are the player's controls defined in config.c.
 ====================================================================
 */
-Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *blocks, SDL_Surface *unknown_preview, char *name, Controls *controls )
+Bowl *bowl_create( int x, int y,
+		int preview_x, int preview_y,
+		int hold_x, int hold_y,
+		SDL_Surface *blocks, SDL_Surface *unknown_preview,
+		char *name, Controls *controls )
 {
     Bowl *bowl = calloc( 1, sizeof( Bowl ) );
     bowl->mute = 0;
@@ -1028,8 +980,10 @@ Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *bloc
     bowl->blocks = blocks;
     bowl->unknown_preview = unknown_preview;
     strcpy( bowl->name, name );
-    bowl->controls = controls;
     bowl->use_figures = ( config.gametype == 2 );
+    /* we use controls only to determine cpu play.
+     * actual controls are determined in tetris_set_bowl_controls */
+    bowl->cpu_player = (controls == 0);
 
     if (config.gametype == 2) {
 	    bowl->firstlevelup_lines = 10;
@@ -1047,21 +1001,29 @@ Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *bloc
 		    bowl->firstlevelup_lines = b;
 	    bowl->level = config.starting_level;
     }
-    //printf("First level %d requires %d lines.\n",
-    //		    	    bowl->level,bowl->firstlevelup_lines);
 
-    bowl->stored_key = -1;
     bowl_reset_contents( bowl );
     bowl->next_block_id = next_blocks[bowl->next_blocks_pos++];
-    delay_set( &bowl->block_hori_delay, 100/*config.hori_delay * 12 + 63*/ );
-    bowl->block_hori_vel = (float)bowl->block_size / bowl->block_hori_delay.limit;
     /* translate config option vert_delay into drop_vel to avoid unnecessary
      * changes in the code. the lower the delay the higher the velocity. */
     bowl->block_drop_vel = 0.6; // fixed to 0.6; 0.8 - config.vert_delay*0.07;
     bowl_set_vert_block_vel( bowl );
     bowl->help_sw = bowl->help_sh = bowl->block_size * 4;
-    bowl->preview_center_sx = preview_x;
-    bowl->preview_center_sy = preview_y;
+    if (!config.preview || preview_x == -1)
+	    bowl->preview = 0;
+    else if (config.modern)
+	    bowl->preview = 3;
+    else
+	    bowl->preview = 1;
+    bowl->preview_sx = preview_x;
+    bowl->preview_sy = preview_y;
+    bowl->preview_sw = bowl->block_size*4;
+    bowl->preview_sh = bowl->block_size + bowl->block_size*3;
+    if (config.modern) {
+	    bowl->preview_sh += 2 * bowl->block_size*3;
+	    /* passed position is for 1 piece so adjust */
+	    bowl->preview_sy -= bowl->block_size*3;
+    }
     bowl_select_next_block( bowl );
     bowl->font = load_fixed_font( "f_small_white.bmp", 32, 96, 8 );
 #ifdef SOUND
@@ -1072,14 +1034,17 @@ Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *bloc
     bowl->wav_excellent = sound_chunk_load( "excellent.wav" );
 #endif
 
-    /* das charge, either class 16/6 or fast 10/3 */
+    /* das charge, either class 16/6 or fast 10/3
     if (config.hyper_das) {
 	    bowl->das_maxcharge = 167;
 	    bowl->das_drop = 50;
     } else {
 	    bowl->das_maxcharge = 267;
 	    bowl->das_drop = 100;
-    }
+    } */
+    /* use configurable das */
+    bowl->das_maxcharge = config.as_delay;
+    bowl->das_drop = config.as_speed;
     bowl->das_charge = 0;
 
     /* lock delay */
@@ -1089,14 +1054,28 @@ Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *bloc
 	    bowl->ldelay_max = 0;
     bowl->ldelay_cur = 0;
 
-    /* stats are only display for demo, normal and figures */
-    if (config.gametype <= 2 || config.gametype <= GAME_TRAINING) {
-	    bowl->stats_x = 30;
-	    bowl->stats_y = 60;
-	    bowl->stats_w = 160;
-	    bowl->stats_h = 280;
+    /* hold settings */
+    if (hold_x == -1 || !config.modern)
+	    bowl->hold_active = 0;
+    else {
+	    bowl->hold_active = 1;
+	    bowl->hold_id = -1;
+	    bowl->hold_used = 0;
+	    bowl->hold_sx = hold_x;
+	    bowl->hold_sy = hold_y;
+	    bowl->hold_sw = bowl->block_size*4;
+	    bowl->hold_sh = bowl->block_size + bowl->block_size*3;
     }
 
+    /* stats are only display for games with one bowl
+     * which is determined in tetris.c:tetris_init()
+     * XXX as it is only one bowl we hard code position here */
+    bowl->show_stats = 0;
+    bowl->stats_x = 30;
+    bowl->stats_y = 60;
+    bowl->stats_w = 160;
+    bowl->stats_h = 280;
+
     return bowl;
 }
 void bowl_delete( Bowl *bowl )
@@ -1122,28 +1101,6 @@ void bowl_delete( Bowl *bowl )
     free( bowl );
 }
 
-/*
-====================================================================
-Check if key belongs to this bowl and store the value for use in
-bowl_update().
-====================================================================
-*/
-void bowl_store_key( Bowl *bowl, int keysym )
-{
-    if ( !bowl->controls ) return; /* CPU handles this bowl */
-    if ( bowl->controls->left  == keysym ) bowl->stored_key = KEY_LEFT;
-    else
-    if ( bowl->controls->right == keysym ) bowl->stored_key = KEY_RIGHT;
-    else
-    if ( bowl->controls->rot_left  == keysym ) bowl->stored_key = KEY_ROT_LEFT;
-    else
-    if ( bowl->controls->rot_right == keysym ) bowl->stored_key = KEY_ROT_RIGHT;
-    else
-    if ( bowl->controls->down  == keysym ) bowl->stored_key = KEY_DOWN;
-    else
-    if ( bowl->controls->drop == keysym ) bowl->stored_key = KEY_DROP;
-}
-
 /*
 ====================================================================
 Hide/show/update all animations handled by a bowl.
@@ -1165,17 +1122,23 @@ void bowl_hide( Bowl *bowl )
         blit_surf();
         add_refresh_rect( bowl->help_sx, bowl->help_sy, bowl->help_sw, bowl->help_sh );
     }
+    /* hold */
+    if ( bowl->hold_active ) {
+	    DEST( sdl.screen, bowl->hold_sx, bowl->hold_sy,
+			    bowl->hold_sw, bowl->hold_sh);
+	    SOURCE( offscreen, bowl->hold_sx, bowl->hold_sy);
+	    blit_surf();
+	    add_refresh_rect(bowl->hold_sx, bowl->hold_sy,
+			    bowl->hold_sw, bowl->hold_sh);
+    }
     /* preview */
-    if ( bowl->preview_center_sx != -1 ) {
-        DEST( sdl.screen, bowl->preview_center_sx - ( bowl->block_size << 1 ), 
-                          bowl->preview_center_sy - ( bowl->block_size << 1 ), 
-                          bowl->block_size << 2, bowl->block_size << 2 );
-        SOURCE( offscreen, bowl->preview_center_sx - ( bowl->block_size << 1 ), 
-                           bowl->preview_center_sy - ( bowl->block_size << 1 ) );
+    if ( bowl->preview ) {
+        DEST( sdl.screen, bowl->preview_sx, bowl->preview_sy,
+        		bowl->preview_sw, bowl->preview_sh);
+        SOURCE( offscreen, bowl->preview_sx, bowl->preview_sy);
         blit_surf();
-        add_refresh_rect( bowl->preview_center_sx - ( bowl->block_size << 1 ), 
-                          bowl->preview_center_sy - ( bowl->block_size << 1 ), 
-                          bowl->block_size << 2, bowl->block_size << 2 );
+        add_refresh_rect(bowl->preview_sx, bowl->preview_sy,
+        		bowl->preview_sw, bowl->preview_sh);
     }
     /* score */
     DEST( sdl.screen, bowl->score_sx, bowl->score_sy, bowl->score_sw, bowl->score_sh );
@@ -1183,7 +1146,7 @@ void bowl_hide( Bowl *bowl )
     blit_surf();
     add_refresh_rect( bowl->score_sx, bowl->score_sy, bowl->score_sw, bowl->score_sh );
     /* stats */
-    if (bowl->stats_w > 0) {
+    if (bowl->show_stats) {
 	    DEST(sdl.screen, bowl->stats_x, bowl->stats_y, bowl->stats_w, bowl->stats_h);
 	    SOURCE(offscreen, bowl->stats_x, bowl->stats_y);
 	    blit_surf();
@@ -1204,21 +1167,10 @@ void bowl_show( Bowl *bowl )
         bowl_draw_contents( bowl );
     }
 
-    /* block&help */
+    /* piece & help */
     if ( !bowl->hide_block ) {
         for ( j = 0; j < 4; j++ ) {
             for ( i = 0; i < 4; i++ ) {
-                if ( block_masks[bowl->next_block_id].mask[block_masks[bowl->next_block_id].rstart][i][j] ) {
-                    /* preview */
-                    if ( config.preview &&  bowl->preview_center_sx != -1 ) {
-                        DEST( sdl.screen, bowl->preview_sx + tile_x, bowl->preview_sy + tile_y, bowl->block_size, bowl->block_size );
-                        SOURCE( bowl->blocks,
-                        	block_masks[bowl->next_block_id].blockid
-				* bowl->block_size, 0 );
-                        blit_surf();
-                        add_refresh_rect( bowl->preview_sx + tile_x, bowl->preview_sy + tile_y, bowl->block_size, bowl->block_size );
-                    }
-                }
                 if ( block_masks[bowl->block.id].mask[bowl->block.rot_id][i][j] ) {
                     /* help */
                     if ( config.modern && bowl->are == 0) {
@@ -1248,16 +1200,56 @@ void bowl_show( Bowl *bowl )
         }
     }
 
-    /* check if question mark must be displayed */
-    if ( bowl->preview_center_sx != -1 && !config.preview ) {
-        DEST( sdl.screen, bowl->preview_center_sx - bowl->unknown_preview->w / 2, 
-                          bowl->preview_center_sy - bowl->unknown_preview->h / 2, 
-                          bowl->unknown_preview->w, bowl->unknown_preview->h );
-        SOURCE( bowl->unknown_preview, 0, 0 );
-        blit_surf();
-        add_refresh_rect( bowl->preview_center_sx - bowl->unknown_preview->w / 2, 
-                          bowl->preview_center_sy - bowl->unknown_preview->h / 2, 
-                          bowl->unknown_preview->w, bowl->unknown_preview->h  );
+    /* hold piece */
+    if (bowl->hold_active) {
+	    /* title */
+	    bowl->font->align = ALIGN_X_CENTER | ALIGN_Y_CENTER;
+	    write_text(bowl->font, sdl.screen,
+			    bowl->hold_sx + bowl->hold_sw/2,
+			    bowl->hold_sy + bowl->block_size/2,
+			    _("Hold"), OPAQUE );
+
+	    if (bowl->hold_id != -1) {
+		    int hw = bowl->block_size*4, hh = bowl->block_size*2;
+		    DEST( sdl.screen, bowl->hold_sx,
+				    bowl->hold_sy + 3*bowl->block_size/2,
+				    hw,hh);
+		    SOURCE( previewpieces, 0, bowl->hold_id * hh );
+		    blit_surf();
+	    }
+
+	    add_refresh_rect(bowl->hold_sx,bowl->hold_sy,bowl->block_size*4,bowl->block_size*3);
+    }
+
+    /* piece preview */
+    if (bowl->preview) {
+	    /* title */
+	    bowl->font->align = ALIGN_X_CENTER | ALIGN_Y_CENTER;
+	    write_text(bowl->font, sdl.screen,
+			    bowl->preview_sx + bowl->preview_sw/2,
+			    bowl->preview_sy + bowl->block_size/2,
+			    _("Next"), OPAQUE );
+
+	    /* first piece */
+	    int pw = bowl->block_size*4, ph = bowl->block_size*2;
+	    DEST( sdl.screen, bowl->preview_sx,
+			    bowl->preview_sy + 3*bowl->block_size/2,
+			    pw,ph);
+	    SOURCE( previewpieces, 0, bowl->next_block_id * ph );
+	    blit_surf();
+
+	    /* only for modern show next two pieces of piece bag */
+	    for (int i = 0; i < bowl->preview-1; i++) {
+		    int pid = next_blocks[bowl->next_blocks_pos+i];
+		    DEST( sdl.screen, bowl->preview_sx,
+				    bowl->preview_sy + 3*bowl->block_size/2
+				    	    + (ph + bowl->block_size)*(i+1),
+				    pw,ph);
+		    SOURCE( previewpieces, 0, pid * ph );
+		    blit_surf();
+	    }
+
+	    add_refresh_rect(bowl->preview_sx,bowl->preview_sy,pw,ph);
     }
 
     /* score, lines, level */
@@ -1269,7 +1261,7 @@ void bowl_show( Bowl *bowl )
     write_text( bowl->font, sdl.screen, bowl->score_sx + bowl->score_sw - 4, bowl->score_sy + bowl->score_sh, aux, OPAQUE );
 
     /* stats */
-    if (bowl->stats_w > 0)
+    if (bowl->show_stats)
 	    bowl_draw_stats(bowl);
 }
 
@@ -1305,7 +1297,7 @@ void bowl_check_lockdelay(Bowl *bowl)
 	}
 }
 
-void bowl_update( Bowl *bowl, int ms, int game_over )
+void bowl_update( Bowl *bowl, int ms, BowlControls *bc, int game_over )
 {
     int old_y = bowl_convert_cury2y(bowl);
     int hori_movement = 0;
@@ -1323,134 +1315,125 @@ void bowl_update( Bowl *bowl, int ms, int game_over )
 	    bowl->are -= ms;
 	    if (bowl->are <= 0) {
 		    bowl->are = 0;
-		    if (bowl->stored_key == KEY_LEFT ||
-				    bowl->stored_key == KEY_RIGHT)
-			    bowl->stored_key = -1; /* kill input to keep das */
-		    if (bowl->stored_key == KEY_DROP)
-			    bowl->stored_key = -1; /* prevent accidental drop */
+		    /* kill input to keep DAS and prevent drops */
+		    memset(bc,0,sizeof(BowlControls));
 		    /* running are means block was inserted. when expired
 		     * remove empty lines and get next block */
 		    bowl_collapse(bowl);
 		    bowl_select_next_block(bowl);
 		    bowl->draw_contents = 1;
+		    bowl->hold_used = 0; /* allow using hold again */
 		    /* for simple modern version charge das if shift pressed */
-		    if (config.modern) {
-			    if ((bowl->controls && keystate[bowl->controls->left]) ||
-					    ( !bowl->controls && bowl->cpu_dest_x < bowl->block.x))
-				    bowl->das_charge = bowl->das_maxcharge;
-			    if ((bowl->controls && keystate[bowl->controls->right]) ||
-					    ( !bowl->controls && bowl->cpu_dest_x > bowl->block.x))
-				    bowl->das_charge = bowl->das_maxcharge;
-		    }
+		    if (config.modern)
+			    bowl->das_charge = bowl->das_maxcharge;
 	    }
     }
 
     /* BLOCK */
     if ( !bowl->hide_block && bowl->are == 0) {
-        /* fake a key event to rotate with cpu */
-        if ( !bowl->controls && bowl->cpu_dest_rot != bowl->block.rot_id ) 
-            if ( delay_timed_out( &bowl->cpu_rot_delay, ms ) )
-                bowl->stored_key = KEY_ROT_LEFT;
-        /* if CPU may drop in one p-cycle set key */
-        if ( bowl_cpu_may_drop( bowl ) )
-            bowl->stored_key = KEY_DROP;
-        /* handle stored key */
-        switch ( bowl->stored_key ) {
-            case KEY_LEFT:
-            case KEY_RIGHT:
-        	    /* try to instantly move and reset das charge.
-        	     * if shifting not possible, charge to the max */
-        	    bowl->das_charge = 0;
-        	    if (bowl->stored_key == KEY_LEFT)
-        		    hori_mod = -1;
-        	    else
-        		    hori_mod = 1;
-       		    if (bowl_check_piece_position(bowl,
-        				    bowl->block.x + hori_mod,
-					    bowl->block.y,
+	    /* handle controls */
+	    if (bc->lshift == CS_DOWN || bc->rshift == CS_DOWN) {
+		    /* try to instantly move and reset das charge.
+		     * if shifting not possible, charge to the max */
+		    bowl->das_charge = 0;
+		    if (bc->lshift == CS_DOWN)
+			    hori_mod = -1;
+		    else
+			    hori_mod = 1;
+		    if (bowl_check_piece_position(bowl,
+				    bowl->block.x + hori_mod,
+				    bowl->block.y,
+				    bowl->block.rot_id) != POSVALID)
+			    bowl->das_charge = bowl->das_maxcharge;
+		    else {
+			    bowl->block.x += hori_mod;
+			    hori_movement = 1;
+			    bowl_check_lockdelay(bowl);
+		    }
+	    } else if (bc->lrot == CS_DOWN || bc->rrot == CS_DOWN) {
+		    /* test if we actually can rotate
+		     * if not shift block if modern and rotate anyways */
+		    if (bc->lrot == CS_DOWN) {
+			    new_rot = bowl->block.rot_id - 1;
+			    if ( new_rot < 0 )
+				    new_rot = 3;
+		    } else {
+			    new_rot = bowl->block.rot_id + 1;
+			    if ( new_rot == 4 )
+				    new_rot = 0;
+		    }
+		    /* the original wall kick rules allow for ridiculous
+		     * moves so we go with a relatable simple version.
+		     * it is tried to move the piece one to the right, one
+		     * to the left and then one up (in this order). if
+		     * any if these single steps fails the rotation fails.
+		     */
+		    hori_mod = 0;
+		    vert_mod = 0;
+		    ret = bowl_check_piece_position(bowl,
+				    bowl->block.x, bowl->block.y, new_rot);
+		    if (ret != POSVALID) {
+			    if (bowl_check_piece_position(bowl,
+					    bowl->block.x+1, bowl->block.y,
+					    new_rot) == POSVALID) {
+				    hori_mod = 1;
+				    ret = POSVALID;
+			    } else if (bowl_check_piece_position(bowl,
+					    bowl->block.x-1, bowl->block.y,
+					    new_rot) == POSVALID) {
+				    hori_mod = -1;
+				    ret = POSVALID;
+			    } else if (bowl_check_piece_position(bowl,
+					    bowl->block.x, bowl->block.y-1,
+					    new_rot) == POSVALID) {
+				    hori_mod = 0;
+				    vert_mod = -1;
+				    ret = POSVALID;
+			    }
+		    }
+		    if (ret == POSVALID && (config.modern ||
+				    (hori_mod == 0 && vert_mod==0))) {
+			    bowl->block.rot_id = new_rot;
+			    bowl->block.x += hori_mod;
+			    bowl->block.y += vert_mod;
+			    if (vert_mod)
+				    bowl->block.cur_y = bowl->block.y*bowl->block_size;
+			    bowl_compute_help_pos(bowl);
+			    bowl_check_lockdelay(bowl);
+		    }
+	    } else if (bc->hdrop == CS_DOWN) {
+		    if (config.gametype == GAME_TRAINING)
+			    bowl_toggle_gravity(bowl);
+		    else {
+			    bowl_drop_block(bowl);
+			    old_y = bowl->block.y-1; /* prevent warning below */
+			    bowl_check_lockdelay(bowl);
+			    if (bowl->ldelay_cur > 0)
+				    bowl->ldelay_cur = 1;
+		    }
+	    } else if (bc->hold == CS_DOWN && bowl->hold_active && !bowl->hold_used) {
+		    /* put current piece to hold, use piece in hold or next block */
+		    bowl->hold_used = 1;
+		    bowl->ldelay_cur = 0; /* reset lock delay if any */
+		    if (bowl->hold_id == -1) {
+			    bowl->hold_id = bowl->block.id;
+			    bowl_select_next_block(bowl);
+		    } else {
+			    int iswap = bowl->block.id;
+			    bowl->block.id = bowl->hold_id;
+			    bowl->hold_id = iswap;
+			    bowl_init_current_piece(bowl, bowl->block.id);
+			    /* opposite to select_next_block init_current_piece
+			     * does not check for game over so we do this here */
+			    if (bowl_check_piece_position(bowl,
+					    bowl->block.x, bowl->block.y,
 					    bowl->block.rot_id) != POSVALID)
-       			    bowl->das_charge = bowl->das_maxcharge;
-       		    else {
-       			    bowl->block.x += hori_mod;
-       			    hori_movement = 1;
-       			    bowl_check_lockdelay(bowl);
-       		    }
-                break;
-            case KEY_ROT_LEFT:
-            case KEY_ROT_RIGHT:
-        	    /* test if we actually can rotate
-        	     * if not shift block if modern and rotate anyways */
-                if (bowl->stored_key == KEY_ROT_LEFT) {
-                    new_rot = bowl->block.rot_id - 1;
-                    if ( new_rot < 0 )
-                        new_rot = 3;
-                } else {
-                    new_rot = bowl->block.rot_id + 1;
-                    if ( new_rot == 4 )
-                        new_rot = 0;
-                }
-                /* the original wall kick rules allow for ridiculous
-                 * moves so we go with a relatable simple version.
-                 * it is tried to move the piece one to the right, one
-                 * to the left and then one up (in this order). if
-                 * any if these single steps fails the rotation fails.
-                 */
-                hori_mod = 0;
-                vert_mod = 0;
-                ret = bowl_check_piece_position(bowl,
-                		bowl->block.x, bowl->block.y, new_rot);
-                if (ret != POSVALID) {
-                	if (bowl_check_piece_position(bowl,
-                			bowl->block.x+1, bowl->block.y,
-                			new_rot) == POSVALID) {
-                		hori_mod = 1;
-                		ret = POSVALID;
-                	} else if (bowl_check_piece_position(bowl,
-                			bowl->block.x-1, bowl->block.y,
-                			new_rot) == POSVALID) {
-                		hori_mod = -1;
-                		ret = POSVALID;
-                	} else if (bowl_check_piece_position(bowl,
-                			bowl->block.x, bowl->block.y-1,
-                			new_rot) == POSVALID) {
-                		hori_mod = 0;
-                		vert_mod = -1;
-                		ret = POSVALID;
-                	}
-                }
-                if (ret == POSVALID && (config.modern ||
-                			(hori_mod == 0 && vert_mod==0))) {
-                	bowl->block.rot_id = new_rot;
-                	bowl->block.x += hori_mod;
-                	bowl->block.y += vert_mod;
-                	if (vert_mod)
-                		bowl->block.cur_y = bowl->block.y*bowl->block_size;
-                        bowl_compute_help_pos(bowl);
-                        bowl_check_lockdelay(bowl);
-                }
-                break;
-            case KEY_DROP:
-                if (config.gametype == GAME_TRAINING)
-                    bowl_toggle_gravity(bowl);
-                else {
-                    bowl_drop_block(bowl);
-                    bowl_check_lockdelay(bowl);
-                    if (bowl->ldelay_cur > 0)
-                	    bowl->ldelay_cur = 1;
-                }
-                break;
-        }
+				    bowl_finish_game(bowl);
+		    }
+	    }
 
         /* update horizontal bowl position */
-        int lshift = 0, rshift = 0;
-        /* check if left or right shift requested */
-        if ((bowl->controls && keystate[bowl->controls->left]) ||
-        		( !bowl->controls && bowl->cpu_dest_x < bowl->block.x))
-        	lshift = 1;
-        if ((bowl->controls && keystate[bowl->controls->right]) ||
-        		( !bowl->controls && bowl->cpu_dest_x > bowl->block.x))
-        	rshift = 1;
-        if (lshift || rshift) {
+        if (bc->lshift == CS_PRESSED || bc->rshift == CS_PRESSED) {
                 /* charge das */
         	if (bowl->das_charge < bowl->das_maxcharge) {
         		bowl->das_charge += ms;
@@ -1459,7 +1442,7 @@ void bowl_update( Bowl *bowl, int ms, int game_over )
         	}
                 /* perform auto shift */
                 if (bowl->das_charge == bowl->das_maxcharge) {
-                	if (lshift)
+                	if (bc->lshift == CS_PRESSED)
                 		hori_mod = -1;
                 	else
                 		hori_mod = 1;
@@ -1487,14 +1470,13 @@ void bowl_update( Bowl *bowl, int ms, int game_over )
         bowl->block.sx = bowl->block.x * bowl->block_size + bowl->sx;
 
         /* update vertical position */
-        if ( !bowl->controls && !bowl->cpu_down )
+        if ( bowl->cpu_player && !bowl->cpu_down )
         	if ( delay_timed_out( &bowl->cpu_delay, ms ) )
         		bowl->cpu_down = 1;
         vy = bowl->block_vert_vel;
-        if ((bowl->controls && keystate[bowl->controls->down]) ||
-        				bowl_cpu_may_move_down(bowl))
+        if (bc->sdrop == CS_PRESSED || bowl_cpu_may_move_down(bowl))
         	if (bowl->block_drop_vel > vy)
-        		if (config.modern || (!lshift && !rshift))
+        		if (config.modern || (bc->lshift!=CS_PRESSED && bc->rshift!=CS_PRESSED))
         			vy = bowl->block_drop_vel;
         if (!bowl->zero_gravity && bowl->ldelay_cur == 0) {
         	bowl->block.cur_y += vy * ms;
@@ -1502,7 +1484,7 @@ void bowl_update( Bowl *bowl, int ms, int game_over )
         }
 
         /* skip lock delay with soft drop */
-        if (bowl->controls && keystate[bowl->controls->down])
+        if (bc->sdrop == CS_PRESSED)
         	if (bowl->ldelay_cur > 0)
         		bowl->ldelay_cur = 1;
 
@@ -1543,9 +1525,6 @@ void bowl_update( Bowl *bowl, int ms, int game_over )
             bowl->block.sy = bowl->block.y * bowl->block_size + bowl->sy;
         else
             bowl->block.sy = (int)bowl->block.cur_y + bowl->sy;
-
-        /* clear stored key */
-        bowl->stored_key = -1;
     }
 
     /* CHECK SPECIAL EVENTS */
@@ -1652,18 +1631,20 @@ void bowl_draw_frames( Bowl *bowl )
     bowl->font->align = ALIGN_X_LEFT | ALIGN_Y_BOTTOM;
     write_text( bowl->font, bkgnd, dx + 4, dy + dh - 4, _("Lines:"), OPAQUE );
     /* preview */
-    if ( bowl->preview_center_sx != -1 )
-        draw_3dframe( bkgnd, 
-                      bowl->preview_center_sx - 2 * bowl->block_size - 2, bowl->preview_center_sy - 2 * bowl->block_size - 2, 
-                      4 * bowl->block_size + 4, 4 * bowl->block_size + 4,
-                      4 );
+    if (bowl->preview)
+	    draw_3dframe( bkgnd, bowl->preview_sx - 2, bowl->preview_sy - 2,
+			    bowl->preview_sw + 4, bowl->preview_sh + 4, 4 );
+    /* hold */
+    if (bowl->hold_active)
+	    draw_3dframe( bkgnd, bowl->hold_sx - 2, bowl->hold_sy - 2,
+			    bowl->hold_sw + 4, bowl->hold_sh + 4, 4 );
     /* part that is updated when redrawing score/level */
     bowl->score_sx = dx + dw / 2 - 36;
     bowl->score_sy = dy + bowl->font->height + 4;
     bowl->score_sw = dw / 2 + 36;
     bowl->score_sh = dh - bowl->font->height - 8;
     /* stats */
-    if (bowl->stats_w > 0)
+    if (bowl->show_stats)
 	    draw_3dframe(bkgnd, bowl->stats_x, bowl->stats_y, bowl->stats_w, bowl->stats_h, 4);
 }
 
@@ -1679,7 +1660,6 @@ void bowl_toggle_pause( Bowl *bowl )
         bowl->hide_block = 0;
         bowl_draw_contents( bowl );
         bowl->paused = 0;
-		bowl->stored_key = -1;
     }
     else {
         /* pause */
