@@ -20,16 +20,22 @@
 
 #include "ltris.h"
 
-enum { 
-    KEY_NONE = -1,
-    KEY_LEFT, 
-    KEY_RIGHT,
-    KEY_ROT_LEFT,
-    KEY_ROT_RIGHT,
-    KEY_DOWN,
-    KEY_DROP
+/* control states */
+enum {
+	CS_RELEASED = 0, /* continuously released */
+	CS_PRESSED, /* continuously pressed */
+	CS_DOWN, /* just pressed down */
+	CS_UP /* just released */
 };
 
+/* bowl controls */
+typedef struct {
+	int lshift, rshift;
+	int lrot, rrot;
+	int sdrop, hdrop;
+	int hold;
+} BowlControls;
+
 typedef struct {
     int sx, sy; /* screen position */
     int sw, sh; /* tile size in pixels */
@@ -59,14 +65,12 @@ typedef struct {
     int sw, sh; /* screen size */
     int w, h; /* measurements in blocks */
     int block_size; /* blocksize in pixels */
-    Controls *controls; /* reacts to these controls */
     int das_charge; /* current charge in ms */
     int das_maxcharge; /* maximum charge in ms */
     int das_drop; /* das charge drop if shifting piece */
     int are; /* in ms, if > 0 next piece is blocked until delay times out */
     int ldelay_max; /* time on collision until piece is inserted */
     int ldelay_cur; /* current value if > 0 */
-    int stored_key; /* key that was stored this programme cycle */
     SDL_Surface *blocks; /* pointer to the block graphics */
     SDL_Surface *unknown_preview; /* if preview's unknown this is displayed */
     char name[32]; /* player's name for this bowl */
@@ -83,12 +87,11 @@ typedef struct {
     Delay add_delay; /* delay until next add action */
     int contents[BOWL_WIDTH][BOWL_HEIGHT]; /* indices of blocks or -1 */
     Block block;/* current block */
-    Delay block_hori_delay; /* horizontal movement delay */
     int next_block_id; /* id of next block */
     int use_same_blocks; /* use global block list? */
     int next_blocks_pos; /* position in tetris next_blocks for 
                             mulitplayer games */
-    float block_vert_vel, block_hori_vel; /* velocity per ms */
+    float block_vert_vel; /* velocity per ms */
     float block_drop_vel;
     int score_sx, score_sy, score_sw, score_sh; /* region with score and lines/level */
     int game_over; /* set if bowl is filled */
@@ -96,8 +99,17 @@ typedef struct {
     int paused;
     int draw_contents; /* set if bowl needs a full redraw next bowl_show() */
     int help_sx, help_sy, help_sw, help_sh; /* position of helping shadow */
-    int preview_center_sx, preview_center_sy; /* preview is centered here if preview_center_x != -1  */
-    int preview_sx, preview_sy; /* actual preview is drawn here */
+
+    int preview; /* 0 = no preview, otherwise number of pieces */
+    int preview_sx, preview_sy; /* preview position */
+    int preview_sw, preview_sh; /* preview size */
+
+    int hold_active; /* whether hold can be used for this bowl */
+    int hold_id; /* block id or -1 if none */
+    int hold_used; /* true if hold was used, gets reset when block is inserted */
+    int hold_sx, hold_sy, hold_sw, hold_sh; /* region with hold block */
+
+    int cpu_player; /* if 1 cpu controlled */
     int cpu_dest_x; /* move block to this position (computed in bowl_select_next_block() */
     int cpu_dest_rot; /* destination rotation */
     int cpu_dest_score; /* AI score */
@@ -113,6 +125,7 @@ typedef struct {
 #endif
 
     /* statistics */
+    int show_stats;
     int stats_x, stats_y, stats_w, stats_h;
     BowlStats stats;
     int drought; /* current drought: pieces since last i piece */
@@ -140,17 +153,13 @@ Create a bowl at screen position x,y. Measurements are the same for
 all bowls. Controls are the player's controls defined in config.c.
 ====================================================================
 */
-Bowl *bowl_create( int x, int y, int preview_x, int preview_y, SDL_Surface *blocks, SDL_Surface *unknown_preview, char *name, Controls *controls );
+Bowl *bowl_create( int x, int y,
+		int preview_x, int preview_y,
+		int hold_x, int hold_y,
+		SDL_Surface *blocks, SDL_Surface *unknown_preview,
+		char *name, Controls *controls );
 void bowl_delete( Bowl *bowl );
 
-/*
-====================================================================
-Check if key belongs to this bowl and store the value for use in
-bowl_update().
-====================================================================
-*/
-void bowl_store_key( Bowl *bowl, int keysym );
-
 /*
 ====================================================================
 Finish game and set game over.
@@ -166,7 +175,7 @@ If game_over only score is updated in bowl_update().
 */
 void bowl_hide( Bowl *bowl );
 void bowl_show( Bowl *bowl );
-void bowl_update( Bowl *bowl, int ms, int game_over );
+void bowl_update( Bowl *bowl, int ms, BowlControls *bc, int game_over );
 
 /*
 ====================================================================
