@@ -18,6 +18,8 @@
 #ifndef __CONFIG_H
 #define __CONFIG_H
 
+#define CONFIG_FILE_NAME "ltris.conf"
+
 /* game type ids */
 enum {
     GAME_DEMO = 0,
@@ -39,6 +41,7 @@ typedef struct {
     int rot_right;
     int down;
     int drop;
+    int hold;
 } Controls;
 
 typedef struct {
@@ -59,6 +62,13 @@ typedef struct {
     Player player1;
     Player player2;
     Player player3;
+    /* gamepad */
+    int gp_enabled;
+    int gp_lrot;
+    int gp_rrot;
+    int gp_hdrop;
+    int gp_pause;
+    int gp_hold;
     /* multiplayer */
     int holes;
     int rand_holes;
@@ -69,7 +79,8 @@ typedef struct {
     int cpu_delay; /* delay in ms that CPU waits before moving down fast */
     int cpu_rot_delay; /* delay between rotative steps */
     /* controls */
-    int	hori_delay;
+    int	as_delay;
+    int	as_speed;
     int vert_delay;
     int pause_key;
     int hyper_das;
@@ -94,12 +105,8 @@ typedef struct {
     int quick_help;
     int visualize; /* compute stats hidden? */
     int keep_bkgnd;
-    int center_preview;
 } Config;
 
-/* config directory name in home directory */
-#define CONFIG_DIR_NAME ".lgames"
-
 /* set config to default */
 void config_reset();
 
